"""
Native accelerators for Parquet encoding and decoding.
"""

from __future__ import absolute_import

cdef extern from "string.h":
    void *memcpy(void *dest, const void *src, size_t n)

from cpython cimport (PyUnicode_AsUTF8String, PyUnicode_DecodeUTF8,
                      PyBytes_CheckExact, PyBytes_FromStringAndSize,
                      PyBytes_GET_SIZE, PyBytes_AS_STRING)

import numpy as np
cimport numpy as np


_obj_dtype = np.dtype('object')


def _to_array(obj):
    """
    Convert *obj* (a ndarray-compatible object, e.g. pandas Series)
    to a true ndarray.
    """
    if not isinstance(obj, np.ndarray):
        try:
            obj = obj.__array__()
        except AttributeError:
            raise TypeError("expected an ndarray-compatible object, got %r"
                            % (type(obj,)))
        assert isinstance(obj, np.ndarray)
    return obj


def _check_1d_object_array(np.ndarray arr):
    if arr.ndim != 1:
        raise TypeError("expected a 1d array")
    if arr.dtype != _obj_dtype:
        raise TypeError("expected an object array")


def array_encode_utf8(inp):
    """
    utf-8 encode all elements of a 1d ndarray of "object" dtype.
    A new ndarray of bytes objects is returned.
    """
    cdef:
        Py_ssize_t i, n
        np.ndarray[object] arr
        np.ndarray[object] result

    arr = _to_array(inp)
    _check_1d_object_array(arr)

    n = len(arr)
    result = np.empty(n, dtype=object)
    for i in range(n):
        # Fast utf-8 encoding, avoiding method call and codec lookup indirection
        result[i] = PyUnicode_AsUTF8String(arr[i])

    return result


def array_decode_utf8(inp):
    """
    utf-8 decode all elements of a 1d ndarray of "object" dtype.
    A new ndarray of unicode objects is returned.
    """
    cdef:
        Py_ssize_t i, n
        np.ndarray[object] arr
        np.ndarray[object] result
        object val

    arr = _to_array(inp)
    _check_1d_object_array(arr)

    n = len(arr)
    result = np.empty(n, dtype=object)
    for i in range(n):
        val = arr[i]
        if not PyBytes_CheckExact(val):
            raise TypeError("expected array of bytes")
        # Fast utf-8 decoding, avoiding method call and codec lookup indirection
        result[i] = PyUnicode_DecodeUTF8(
            PyBytes_AS_STRING(val),
            PyBytes_GET_SIZE(val),
            NULL,   # errors
            )

    return result


def pack_byte_array(list items):
    """
    Pack a variable length byte array column.
    A bytes object is returned.
    """
    cdef:
        Py_ssize_t i, n, itemlen, total_size
        unsigned char *start
        unsigned char *data
        object val, out

    # Strategy: compute the total output size and allocate it in one go.
    n = len(items)
    total_size = 0
    for i in range(n):
        val = items[i]
        if not PyBytes_CheckExact(val):
            raise TypeError("expected list of bytes")
        total_size += 4 + PyBytes_GET_SIZE(val)

    out = PyBytes_FromStringAndSize(NULL, total_size)
    start = data = <unsigned char *> PyBytes_AS_STRING(out)

    # Copy data to output.
    for i in range(n):
        val = items[i]
        # `itemlen` should be >= 0, so no signed extension issues
        itemlen = PyBytes_GET_SIZE(val)
        data[0] = itemlen & 0xff
        data[1] = (itemlen >> 8) & 0xff
        data[2] = (itemlen >> 16) & 0xff
        data[3] = (itemlen >> 24) & 0xff
        data += 4
        memcpy(data, PyBytes_AS_STRING(val), itemlen)
        data += itemlen

    assert (data - start) == total_size
    return out


def unpack_byte_array(bytes raw_bytes, Py_ssize_t n):
    """
    Unpack a variable length byte array column.
    A list of bytes objects is returned.  RuntimeError is raised
    if *raw_bytes* contents don't exactly match *n*.
    """
    cdef:
        Py_ssize_t i, itemlen, remaining
        unsigned char *start
        unsigned char *data
        list out

    start = data = <unsigned char *> PyBytes_AS_STRING(raw_bytes)
    remaining = PyBytes_GET_SIZE(raw_bytes)
    out = [None] * n

    for i in range(n):
        remaining -= 4
        # It is required to check this inside the loop to avoid
        # out of bounds array accesses.
        if remaining < 0:
            raise RuntimeError("Ran out of input")
        itemlen = (data[0] + (data[1] << 8) +
                   (data[2] << 16) + (data[3] << 24))
        data += 4

        remaining -= itemlen
        if remaining < 0:
            raise RuntimeError("Ran out of input")
        out[i] = PyBytes_FromStringAndSize(<char *> data, itemlen)
        data += itemlen

    return out



"""setup.py - build script for parquet-python."""

import os
import sys
try:
    from setuptools import setup, Extension
    from setuptools.command.build_ext import build_ext as _build_ext
except ImportError:
    from distutils.core import setup, Extension
    from distutils.command.build_ext import build_ext as _build_ext

# Kudos to https://stackoverflow.com/questions/19919905/how-to-bootstrap-numpy-installation-in-setup-py/21621689
class build_ext(_build_ext):
    def finalize_options(self):
        if sys.version_info[0] >= 3:
            import builtins
        else:
            import __builtin__ as builtins
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        builtins.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())

allowed = ('--help-commands', '--version', 'egg_info', 'clean')
if len(sys.argv) >= 2 and ('--help' in sys.argv[1:] or sys.argv[1] in allowed):
    # NumPy and cython are not required for these actions. They must succeed
    # so pip can install fastparquet when these requirements are not available.
    extra = {}
else:
    modules_to_build = {
        'fastparquet.speedups': ['fastparquet/speedups.pyx']
    }
    try:
        from Cython.Build import cythonize
        def fix_exts(sources):
            return sources
    except ImportError:
        def cythonize(modules):
            return modules
        def fix_exts(sources):
            return [s.replace('.pyx', '.c') for s in sources]

    modules = [
        Extension(mod, fix_exts(sources))
        for mod, sources in modules_to_build.items()]

    extra = {'ext_modules': cythonize(modules)}

install_requires = open('requirements.txt').read().strip().split('\n')

# add pytest-runner from setup_requires
install_requires.append('pytest-runner')

setup(
    name='fastparquet',
    version='0.1.6',
    description='Python support for Parquet file format',
    author='Martin Durant',
    author_email='mdurant@continuum.io',
    url='https://github.com/dask/fastparquet/',
    license='Apache License 2.0',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    packages=['fastparquet'],
    cmdclass={'build_ext': build_ext},
    install_requires=install_requires,
    # remove wierd setup_requires to get rid of easy_install
    # setup_requires=[
    #    'pytest-runner',
    #    [p for p in install_requires if p.startswith('numpy')][0]
    #],
    tests_require=[
        'pytest',
        'python-snappy',
        'lz4 >= 0.19.1',
        'zstandard',
    ],
    long_description=(open('README.rst').read() if os.path.exists('README.rst')
                      else ''),
    package_data={'fastparquet': ['*.thrift']},
    include_package_data=True,
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*,",
    **extra
)
